@@ -1,16 +1,17 @@
 #!bin/env python3
 
-import re
-import os
 import json
-import yaml
-import sys
+import os
 from pathlib import Path
+import re
+import sys
+
+import yaml
 
 from slalom.dataops import env, io, jobs
 from logless import logged, logged_block, get_logger
 
-BASE_DOCKER_REPO = "aaronsteers/tapdance"  # TODO: Revert to dataopstk/tapdance
+BASE_DOCKER_REPO = "dataopstk/tapdance"
 SINGER_PLUGINS_INDEX = os.environ.get("SINGER_PLUGINS_INDEX", "./singer_index.yml")
 VENV_ROOT = "/venv"
 INSTALL_ROOT = "/usr/bin"
@@ -18,7 +19,7 @@ _ROOT_DIR = "/projects/my-project"
 # _ROOT_DIR = "."
 
 # These plugins will attempt to scrape and pass along AWs Credentials from the local environment.
-S3_TARGET_IDS = ["S3-CSV"]
+S3_TARGET_IDS = ["S3-CSV", "REDSHIFT"]
 
 logging = get_logger("tapdance")
 
@@ -51,6 +52,15 @@ def _get_taps_dir(override=None):
     return io.make_local(taps_dir)  # if remote path provided, download locally
 
 
+def _get_state_file_path():
+    result = os.environ.get("TAP_STATE_FILE", None)
+    if not result:
+        logging.warning(
+            "Could not locate env var 'TAP_STATE_FILE'. State may not be maintained."
+        )
+    return result
+
+
 def _get_catalog_output_dir(tap_name):
     result = f"{_get_scratch_dir()}/taps/{tap_name}-catalog"
     io.create_folder(result)
@@ -324,7 +334,9 @@ def sync(
     for table in list_of_tables:
         # Call each tap independently so that table state files are kept separate:
         tmp_catalog_file = f"{catalog_dir}/{tap_name}-{table}-catalog.json"
-        table_state_file = state_file or f"{catalog_dir}/{table}-state.json"
+        table_state_file = (
+            state_file or _get_state_file_path() or f"{catalog_dir}/{table}-state.json"
+        )
         _create_single_table_catalog(
             tap_name=tap_name,
             table_name=table,
@@ -342,15 +354,16 @@ def sync(
         )
 
 
-def build_all_images(push: bool = False, pre: bool = False):
+def build_all_images(push: bool = False, pre: bool = False, ignore_cache: bool = False):
     """
     Build all images.
 
     :param push: Push images after building
     :param pre: Create and publish pre-release builds
+    :param ignore_cache: True to build images without cached image layers. (default: {False})
     """
-    _build_all_standalone(push=push, pre=pre)
-    _build_all_composite(push=push, pre=pre)
+    _build_all_standalone(push=push, pre=pre, ignore_cache=ignore_cache)
+    _build_all_composite(push=push, pre=pre, ignore_cache=ignore_cache)
 
 
 def build_image(
@@ -358,6 +371,7 @@ def build_image(
     target_alias: str = None,
     push: bool = False,
     pre: bool = False,
+    ignore_cache: bool = False,
 ):
     """
     Build a single image. If tap and target are both provided, any required upstream images
@@ -370,14 +384,28 @@ def build_image(
         target_alias {str} -- Optional. The name of the target (without the `target-` prefix).
         push {bool} -- True to push images to image repository after build. (default: {False})
         pre {bool} -- True to use and create prelease versions. (default: {False})
+        ignore_cache {bool} -- True to build images without cached image layers. (default: {False})
     """
     name, source, alias = _get_plugin_info(f"tap-{tap_or_plugin_alias}")
-    _build_plugin_image(name, source=source, alias=alias, push=push, pre=pre)
+    _build_plugin_image(
+        name, source=source, alias=alias, push=push, pre=pre, ignore_cache=ignore_cache
+    )
     if target_alias:
         name, source, alias = _get_plugin_info(f"target-{target_alias}")
-        _build_plugin_image(name, source=source, alias=alias, push=push, pre=pre)
+        _build_plugin_image(
+            name,
+            source=source,
+            alias=alias,
+            push=push,
+            pre=pre,
+            ignore_cache=ignore_cache,
+        )
         _build_composite_image(
-            tap_alias=tap_or_plugin_alias, target_alias=target_alias, push=push, pre=pre
+            tap_alias=tap_or_plugin_alias,
+            target_alias=target_alias,
+            push=push,
+            pre=pre,
+            ignore_cache=ignore_cache,
         )
 
 
@@ -401,17 +429,43 @@ def _get_customized_target_config_file(
     config_defaults = json.loads(io.get_text_file_contents(target_config_file))
     new_config = config_defaults.copy()
     if target_name.upper() in S3_TARGET_IDS:
-        logging.info("Scanning for AWS creds for target 'target-{target_name}'")
-        for secret in ["aws_access_key_id", "aws_secret_access_key"]:
-            if secret not in config_defaults:
-                if secret in os.environ:
-                    logging.info(f"Found env var {secret} for 'target-{target_name}'")
-                    new_config[secret] = os.environ[secret]
-                if secret.upper() in os.environ:
-                    logging.info(
-                        f"Found env var {secret.upper()} for 'target-{target_name}'"
-                    )
-                    new_config[secret] = os.environ[secret.upper()]
+        if (
+            "aws_access_key_id" in config_defaults
+            and "aws_secret_access_key" in config_defaults
+        ):
+            logging.info("AWS creds captured from 'target-{target_name}' config")
+        else:
+            logging.info("Scanning for AWS creds for target 'target-{target_name}'...")
+            (
+                aws_access_key_id,
+                aws_secret_access_key,
+                aws_session_token,
+            ) = io.parse_aws_creds()
+            if aws_access_key_id and aws_secret_access_key:
+                logging.info(
+                    f"Passing 'aws_access_key_id' and 'aws_secret_access_key' "
+                    f"credentials to 'target-{target_name}'"
+                )
+                new_config["aws_access_key_id"] = aws_access_key_id
+                new_config["aws_secret_access_key"] = aws_secret_access_key
+            else:
+                logging.warning(
+                    f"Could not find 'aws_access_key_id' and 'aws_secret_access_key' "
+                    f"credentials for 'target-{target_name}'"
+                )
+            if aws_session_token:
+                logging.info(f"Passing 'aws_session_token' to 'target-{target_name}'")
+                new_config["aws_session_token"] = aws_session_token
+    new_config = _replace_placeholders(
+        new_config, tap_name, table_name, tap_version_num
+    )
+    new_file_path = target_config_file.replace(".json", f"-{table_name}.json")
+    io.create_text_file(new_file_path, json.dumps(new_config))
+    return new_file_path
+
+
+def _replace_placeholders(config_dict, tap_name, table_name, tap_version_num):
+    new_config = config_dict.copy()
     for setting_name in new_config.keys():
         for param, replacement_value in {
             "tap": tap_name,
@@ -427,9 +481,7 @@ def _get_customized_target_config_file(
                 new_config[setting_name] = new_config[setting_name].replace(
                     search_key, replacement_value
                 )
-    new_file_path = target_config_file.replace(".json", f"{table_name}.json")
-    io.create_text_file(new_file_path, json.dumps(new_config))
-    return new_file_path
+    return new_config
 
 
 def _sync_one_table(
@@ -442,20 +494,44 @@ def _sync_one_table(
     table_state_file: str,
 ):
     tap_cmd = f"tap-{tap_name} --config {config_file} --catalog {table_catalog_file}"
+    tap_version_num = "1"  # Not yet supported
+    table_state_file = _replace_placeholders(
+        {"table_state_file": table_state_file}, tap_name, table_name, tap_version_num
+    )["table_state_file"]
+
     if io.file_exists(table_state_file):
-        local_state_file = io.make_local(table_state_file)
-        tap_cmd = f"{tap_cmd} --state {local_state_file}"
+        if io.is_local(table_state_file):
+            local_state_file = table_state_file
+        else:
+            local_state_file = os.path.join(
+                io.get_scratch_dir(), os.path.basename(table_state_file)
+            )
+            io.download_file(table_state_file, local_state_file)
+        tap_cmd += f" --state {local_state_file}"
+    else:
+        local_state_file = os.path.join(
+            io.get_temp_dir(), f"{tap_name}-{table_name}-state.json"
+        )
     tmp_target_config = _get_customized_target_config_file(
-        target_name, target_config_file, tap_name=tap_name, table_name=table_name
+        target_name,
+        target_config_file,
+        tap_name=tap_name,
+        table_name=table_name,
+        tap_version_num=tap_version_num,
     )
     sync_cmd = (
         f"{tap_cmd} | target-{target_name} --config {tmp_target_config} "
-        f">> {table_state_file}"
+        f">> {local_state_file}"
     )
     jobs.run_command(sync_cmd)
     # TODO: decide whether trimming to only the final line is necessary
     # tail -1 state.json > state.json.tmp && mv state.json.tmp state.json
-    if local_state_file != table_state_file:
+    if not io.file_exists(local_state_file):
+        logging.warning(
+            f"State file does not exist at path '{local_state_file}'. Skipping upload. "
+            f"This can be caused by having no data, or no new data, in the source table."
+        )
+    elif local_state_file != table_state_file:
         io.upload_file(local_state_file, table_state_file)
 
 
@@ -704,7 +780,9 @@ def _get_plugins_list(plugins_index=None):
     return list_of_tuples
 
 
-def _build_all_standalone(source_image=None, plugins_index=None, push=False, pre=False):
+def _build_all_standalone(
+    source_image=None, plugins_index=None, push=False, pre=False, ignore_cache=False
+):
     plugins = _get_plugins_list(plugins_index)
     created_images = []
     for name, source, alias in plugins:
@@ -716,6 +794,7 @@ def _build_all_standalone(source_image=None, plugins_index=None, push=False, pre
                 source_image=source_image,
                 push=push,
                 pre=pre,
+                ignore_cache=ignore_cache,
             )
         )
     return created_images
@@ -729,7 +808,9 @@ def _get_plugin_info(plugin_id, plugins_index=None):
     raise ValueError(f"Could not file a plugin called '{plugin_id}'")
 
 
-def _build_all_composite(source_image=None, plugins_index=None, push=False, pre=False):
+def _build_all_composite(
+    source_image=None, plugins_index=None, push=False, pre=False, ignore_cache=False
+):
     plugins = _get_plugins_list(plugins_index)
     created_images = []
     for tap_name, tap_source, tap_alias in plugins:
@@ -738,18 +819,32 @@ def _build_all_composite(source_image=None, plugins_index=None, push=False, pre=
             target_alias = target_alias or target_name
             if tap_alias.startswith("tap-") and target_alias.startswith("target-"):
                 created_images.append(
-                    _build_composite_image(tap_alias, target_alias, push=push, pre=pre)
+                    _build_composite_image(
+                        tap_alias,
+                        target_alias,
+                        push=push,
+                        pre=pre,
+                        ignore_cache=ignore_cache,
+                    )
                 )
     return created_images
 
 
 def _build_plugin_image(
-    plugin_name, source, alias, source_image=None, push=False, pre=False
+    plugin_name,
+    source,
+    alias,
+    source_image=None,
+    push=False,
+    pre=False,
+    ignore_cache=False,
 ):
     source = source or plugin_name
     alias = alias or plugin_name
     image_name = f"{BASE_DOCKER_REPO}:{alias}"
     build_cmd = f"docker build"
+    if ignore_cache:
+        build_cmd += " --no-cache"
     if source_image:
         build_cmd += f" --build-arg source_image={source_image}"
     if pre:
@@ -769,13 +864,17 @@ def _build_plugin_image(
     return image_name
 
 
-def _build_composite_image(tap_alias, target_alias, push=False, pre=False):
+def _build_composite_image(
+    tap_alias, target_alias, push=False, pre=False, ignore_cache=False
+):
     if tap_alias.startswith("tap-"):
         tap_alias = tap_alias.replace("tap-", "", 1)
     if target_alias.startswith("target-"):
         target_alias = target_alias.replace("target-", "", 1)
     image_name = f"{BASE_DOCKER_REPO}:{tap_alias}-to-{target_alias}"
     build_cmd = f"docker build"
+    if ignore_cache:
+        build_cmd += " --no-cache"
     if pre:
         build_cmd += " --build-arg source_image_suffix=--pre"
         image_name += "--pre"
